/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.indikator;

import java.util.Collection;
import java.util.List;

import de.bsvrz.ibv.anastau.bestimmungstau.stau.StauObjektVerwaltung;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.StoerfallIndikator;

/**
 * Schnittstelle zum Zugriff auf eine Fahrtrichtung, d.h. die Verkettung von
 * Störfallindikatoren innerhalb einer Straße, über die die Stauobjektbestimmung
 * möglich ist.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public interface IFahrtRichtung {

	/**
	 * Die Funktion fügt der Fahrtrichtung einen Indikator hinzu, die Positionen
	 * der Indikatoren werden als ungültig erklärt und müssen gegebenenfalls neu
	 * berechnet werden.
	 *
	 * @param index
	 *            die Position an der der Indikator eingefügt werden soll.
	 * @param indikator
	 *            der einzufügende Indikator
	 */
	void add(int index, ISobIndikator indikator);

	/**
	 * Die Funktion fügt der Fahrtrichtung einen Indikator hinzu, die Positionen
	 * der Indikatoren werden als ungültig erklärt und müssen gegebenenfalls neu
	 * berechnet werden.
	 *
	 * @param indikator
	 *            der Indikator
	 * @return Indikator eingefügt ?
	 */
	boolean add(ISobIndikator indikator);

	/**
	 * Die Funktion fügt der Fahrtrichtung eine Liste von Indikatoren hinzu, die
	 * Positionen der Indikatoren werden als ungültig erklärt und müssen
	 * gegebenenfalls neu berechnet werden.
	 *
	 * @param indikatoren
	 *            die einzufügenden Indikatoren
	 * @return Indikatoren eingefügt ?
	 */
	boolean addAll(Collection<? extends ISobIndikator> indikatoren);

	/**
	 * Die Funktion fügt der Fahrtrichtung eine Liste von Indikatoren hinzu, die
	 * Positionen der Indikatoren werden als ungültig erklärt und müssen
	 * gegebenenfalls neu berechnet werden.
	 *
	 * @param index
	 *            die Position, an der die Indikatoren eingefügt werden sollen
	 * @param indikatoren
	 *            die einzufügenden Indikatoren
	 * @return Indikatoren eingefügt ?
	 */
	boolean addAll(int index, Collection<? extends ISobIndikator> indikatoren);

	/**
	 * ermittelt die Staus für die Fahrtrichtung.<br>
	 * Alle geänderten Störfallindikatoren werden gesucht. Ist der neue Zustand
	 * eines Indikators "STAU", wird ein in der Nähe liegender Stau angepasst
	 * bzw. ein neues Stauobjekt erzeugt.
	 *
	 * @param zeitpunkt
	 *            der Zeitpunkt, zu dem die Aktualisierung erfolgt
	 */
	void aktualisiereStaus(long zeitpunkt);

	/**
	 * berechnet die Positionen aller Störfallindikatoren der Fahrtrichtung.
	 */
	void berechnePositionen();

	/**
	 * entfernt alle Störfallindikatoren von der Fahrtrichtung.
	 */
	void clear();

	/**
	 * liefert die Elemente, die in der Fahrtrichtung miteinander verbunden
	 * sind.
	 *
	 * @return die Liste der Indikatoren
	 */
	List<ISobIndikator> getElemente();

	/**
	 * liefert einen ausführlichen Namen für die Fahrtrichtung.<br>
	 * Der Name setzt sich zusammen aus dem Name, der dem Konstruktor des
	 * Objekts übergeben wurde, sowie dem Name des Störfallindikator, mit dem
	 * Suche nach der Verkettung der Indikatoren begonnen hat.
	 *
	 * @return den Name
	 */
	String getFullName();

	/**
	 * liefert die maximale zulässige Länge um die ein Stauobjekt im angegebenen
	 * Bereich durch die Prognose verlängert werden kann.
	 *
	 * @param start
	 *            der Anfang des betracheteten Bereiches
	 * @param ende
	 *            das Ende des betracheteten Bereiches
	 * @return die Länge
	 */
	double getMaxPrognoseLaenge(double start, double ende);

	/**
	 * liefert den Namen der Fahrtrichtung.
	 *
	 * @return den Namen
	 */
	String getName();

	/**
	 * liefert den Störfallindikator, mit der die Suche nach der Verkettung
	 * begonnen wurde.
	 *
	 * @return den Indikator
	 */
	StoerfallIndikator getStart();

	/**
	 * liefert das Objekt, mit dem die Staus für die Fahrtrichtung verwaltet
	 * werden.
	 *
	 * @return die Staiverwaltung der Fahrtrichtung
	 */
	StauObjektVerwaltung getStaus();

	/**
	 * liefert die Strasse, zu der die Fahrtrichtung gehört.
	 *
	 * @return die Strasse
	 */
	ISobStrasse getStrasse();

	/**
	 * Die Funktion entfernt einen Störfallindikator von der Fahrtrichtung. Die
	 * Positionen der Indikatoren werden als ungültig markiert.
	 *
	 * @param index
	 *            die Position, an der der Indikator entfernt werden soll
	 * @return den entfernten Indikator
	 */
	ISobIndikator remove(int index);

	/**
	 * Die Funktion entfernt einen Störfallindikator von der Fahrtrichtung. Die
	 * Positionen der Indikatoren werden als ungültig markiert.
	 *
	 * @param indikator
	 *            der Indikator, der entfernt werden soll
	 * @return Indikator konnte entfernt werden ?
	 */
	boolean remove(Object indikator);

	/**
	 * Die Funktion setzt einen Störfallindikator in der Liste der Indikatoren.
	 * Die Positionen der Indikatoren werden als ungültig markiert.
	 *
	 * @param index
	 *            die Position, an der der neue Störfallindikator gesetzt werden
	 *            soll
	 * @param indikator
	 *            der Indikator
	 * @return der zuvor an dieser Position befindliche Indikator
	 */
	ISobIndikator set(int index, ISobIndikator indikator);
}
