/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.indikator;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;
import de.bsvrz.ibv.anastau.bestimmungstau.parameter.ISobParameterObserver;
import de.bsvrz.ibv.anastau.migration.modell.SystemObjekt;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.AeusseresStrassenSegment;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.InneresStrassenSegment;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.StoerfallIndikator;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.Strasse;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.StrassenKnoten;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.StrassenTeilSegment;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Repräsentation einer Strasse innerhalb der Stauobjektbestimmung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SobStrasse implements ISobStrasse {

    /**
     * Logger für Debugausgaben.
     */
    private static final Debug LOGGER = Debug.getLogger();

    /**
     * Liste der {@link FahrtRichtung}en, durch die die Straße beschrieben wird.
     */
    private List<FahrtRichtung> richtungen;

    /**
     * das VerkehrsmodellNetz, zu dem die Strasse gehört.
     */
    private final VerkehrModellNetz netz;

    /**
     * die zugeordnete Strasse aus der Datenverteilerkonfiguration.
     */
    private final Strasse strasse;

    /**
     * die Verwaltung für die Parameter der Stauobjektbestimmung.
     */
    private final ISobParameterObserver parameter;

    private StauverlaufsAnalyse applikation;

    /**
     * Konstruktor.<br>
     * Die Funktion erzeugt eine INstanz eines Straßenobjekts zur
     * Stauobjektbestimmung. Aus dem übergebenen VerkehrsmodellNetz werden alle der
     * angegebenen Straße zuzuordnenden STörfallindikatoren ermittelt und in
     * {@link FahrtRichtung}en miteinander verbunden.
     *
     * @param applikation die Applikation, die die Stauverlaufsanalyse durchführt
     * @param netz        das verwendete {@link VerkehrModellNetz}
     * @param strasse     die verwendete {@link Strasse}
     * @param parameter   das Objekt zur Verwaltung der Parameter zur
     *                    Stauobjektbestimmung
     */
    public SobStrasse(StauverlaufsAnalyse applikation, final VerkehrModellNetz netz, final Strasse strasse,
            final ISobParameterObserver parameter) {
        this.applikation = applikation;
        this.netz = netz;
        this.strasse = strasse;
        this.parameter = parameter;
        getFahrtRichtungen();
    }

    /**
     * erzeugt einen Indikator für die Stauobjektbestimmung.
     *
     * @param richtung  die {@link FahrtRichtung} für die der Indikator verwendet
     *                  werden soll.
     * @param indikator der Indikator aus der Datenverteilerkonfiguration
     * @return der erzeugte Indikator
     */
    private ISobIndikator erzeugeIndikator(final FahrtRichtung richtung, final StoerfallIndikator indikator) {
        ISobIndikator result = null;

        if (indikator instanceof AeusseresStrassenSegment) {
            result = new SobAeusseresStrassenSegment(richtung, (AeusseresStrassenSegment) indikator);
        } else if (indikator instanceof InneresStrassenSegment) {
            result = new SobInneresStrassenSegment(richtung, (InneresStrassenSegment) indikator);
        } else if (indikator instanceof StrassenTeilSegment) {
            result = new SobStrassenTeilSegment(richtung, (StrassenTeilSegment) indikator);
        } else if (indikator instanceof StrassenKnoten) {
            result = new SobStrassenKnoten(richtung, (StrassenKnoten) indikator);
        } else if (indikator instanceof MessQuerschnittAllgemein) {
            result = new SobMessQuerschnitt(richtung, (MessQuerschnittAllgemein) indikator);
        }

        return result;
    }

    @Override
    public final Collection<FahrtRichtung> getFahrtRichtungen() {
        if (richtungen == null) {
            richtungen = new ArrayList<>();

            final List<AeusseresStrassenSegment> assListe = new ArrayList<>(netz.getAssListe(strasse));
            // die Liste der Strassensegmente wird alphabetisch sortiert, um
            // immer eine definierte Reihenfolge der Liste zu erhalten.
            Collections.sort(assListe, new SystemObjekt.PidComparator());

            int strZugIdx = 1;
            while (assListe.size() > 0) {

                final AeusseresStrassenSegment ass = assListe.get(0);
                final FahrtRichtung strZug = new FahrtRichtung(applikation, this, "Richtung " + (strZugIdx++), ass);
                richtungen.add(strZug);

                SobStrasse.LOGGER.finest("Strassenzug gegen Fahrtrichtung durchsuchen: " + strZug);
                // entgegengesetzt der Fahrtrichtung suchen
                sucheGegenFahrtrichtung(assListe, strZug, ass);

                SobStrasse.LOGGER.finest("Strassenzug in Fahrtrichtung durchsuchen: " + strZug);
                // in Fahrtrichtung suchen
                sucheInFahrtrichtung(assListe, strZug, ass);
            }

        }
        return richtungen;
    }

    @Override
    public final VerkehrModellNetz getNetz() {
        return netz;
    }

    @Override
    public final ISobParameterObserver getParameter() {
        return parameter;
    }

    @Override
    public final Strasse getStrasse() {
        return strasse;
    }

    /**
     * sucht alle Indikatoren entgegengesetzt der Fahrtrichtung ausgehend vom
     * übergebenen Straßensegment.
     *
     * @param assListe die Liste der verfügbaren äußeren Straßensegmente
     * @param strZug   die Fahrtrichtung, die gebildet wird
     * @param ass      das Straßensegment, an dem die Suche beginnt.
     */
    private void sucheGegenFahrtrichtung(final List<AeusseresStrassenSegment> assListe, final FahrtRichtung strZug,
            final AeusseresStrassenSegment ass) {

        AeusseresStrassenSegment letztesSegment = ass;
        ISobIndikator next = new SobAeusseresStrassenSegment(strZug, letztesSegment);

        while (next != null) {

            SobStrasse.LOGGER.finest("Aktueller Indikator: " + next.getIndikator().getName());

            if (strZug.contains(next)) {
                String message = MessageFormat.format("Zirkuläre Verknüpfungen in der Strasse: {0}",
                        strasse.getSystemObject());
                applikation.getBetriebsmeldungssender().warnung(strasse.getPid(), message);
                LOGGER.warning(() -> message);
                break;
            }

            strZug.add(next);
            if (next instanceof SobAeusseresStrassenSegment) {
                letztesSegment = ((SobAeusseresStrassenSegment) next).getIndikator();
                assListe.remove(letztesSegment);
                for (final StrassenTeilSegment teilSegment : letztesSegment.getStrassenTeilSegmente()) {
                    // strZug.add(index, erzeugeIndikator(strZug, teilSegment));
                    strZug.add(erzeugeIndikator(strZug, teilSegment));
                }
                for (final MessQuerschnittAllgemein mq : letztesSegment.getMessquerschnitte()) {
                    final StrassenTeilSegment sts = mq.getStrassenTeilSegment();
                    int mqIndex = strZug.indexOf(next) + 1;
                    if (sts != null) {
                        mqIndex = Math.max(0, strZug.findOffset(sts) + 1);
                    } else {
                        String message = MessageFormat
                                .format("MQ {0} kann keinem Strassenteilsegment zugeordnet werden", mq);
                        applikation.getBetriebsmeldungssender().warnung(mq.getPid(), message);
                        LOGGER.warning(() -> message);
                    }
                    strZug.add(mqIndex, erzeugeIndikator(strZug, mq));
                }

                next = erzeugeIndikator(strZug, letztesSegment.getVonKnoten());
            } else if (next instanceof SobStrassenKnoten) {
                final SobStrassenKnoten knoten = (SobStrassenKnoten) next;
                next = erzeugeIndikator(strZug,
                        ((SobStrassenKnoten) next).getIndikator().getInnereVerbindungDavor(letztesSegment));
                knoten.setIss(next);
            } else if (next instanceof SobInneresStrassenSegment) {
                final int index = strZug.indexOf(next) + 1;
                for (final StrassenTeilSegment teilSegment : ((SobInneresStrassenSegment) next).getIndikator()
                        .getStrassenTeilSegmente()) {
                    strZug.add(index, erzeugeIndikator(strZug, teilSegment));
                }
                // if (StauverlaufsAnalyse.getOptionen().benutzeIssMq()) {
                for (final MessQuerschnittAllgemein mq : ((SobInneresStrassenSegment) next).getIndikator()
                        .getMessquerschnitte()) {
                    final StrassenTeilSegment sts = mq.getStrassenTeilSegment();
                    int mqIndex = strZug.indexOf(next) + 1;
                    if (sts != null) {
                        mqIndex = Math.max(0, strZug.findOffset(sts) + 1);
                    } else {
                        String message = MessageFormat
                                .format("MQ {0} kann keinem Strassenteilsegment zugeordnet werden", mq);
                        applikation.getBetriebsmeldungssender().warnung(mq.getPid(), message);
                        LOGGER.warning(() -> message);
                    }
                    strZug.add(mqIndex, erzeugeIndikator(strZug, mq));
                }
                // }
                AeusseresStrassenSegment vonSegment = ((SobInneresStrassenSegment) next).getIndikator().getVonSegment();
                if (!assListe.contains(vonSegment)) {
                    vonSegment = null;
                }
                next = erzeugeIndikator(strZug, vonSegment);
            }
        }
    }

    /**
     * sucht alle Indikatoren in der Fahrtrichtung ausgehend vom übergebenen
     * Straßensegment.
     *
     * @param assListe die Liste der verfügbaren äußeren Straßensegmente
     * @param strZug   die Fahrtrichtung, die gebildet wird
     * @param ass      das Straßensegment, an dem die Suche beginnt.
     */
    private void sucheInFahrtrichtung(final List<AeusseresStrassenSegment> assListe, final FahrtRichtung strZug,
            final AeusseresStrassenSegment ass) {

        AeusseresStrassenSegment letztesSegment = ass;
        ISobIndikator next = erzeugeIndikator(strZug, ass.getNachKnoten());
        while (next != null) {

            SobStrasse.LOGGER.finest("Aktueller Indikator in Fahrtrichtung: " + next.getIndikator().getName());

            if (strZug.contains(next)) {
                String message = MessageFormat.format("Zirkuläre Verknüpfungen in der Strasse: {0}",
                        strasse.getSystemObject());
                applikation.getBetriebsmeldungssender().warnung(strasse.getPid(), message);
                LOGGER.warning(() -> message);
                break;
            }
            if (next instanceof SobInneresStrassenSegment) {
                strZug.add(1, next);
            } else {
                strZug.add(0, next);
            }
            if (next instanceof SobAeusseresStrassenSegment) {

                letztesSegment = ((SobAeusseresStrassenSegment) next).getIndikator();
                assListe.remove(letztesSegment);
                final int idx = strZug.indexOf(next) + 1;
                for (final StrassenTeilSegment teilSegment : letztesSegment.getStrassenTeilSegmente()) {
                    strZug.add(idx, erzeugeIndikator(strZug, teilSegment));
                }

                for (final MessQuerschnittAllgemein mq : letztesSegment.getMessquerschnitte()) {
                    final StrassenTeilSegment sts = mq.getStrassenTeilSegment();
                    int mqIndex = strZug.indexOf(next) + 1;
                    if (sts != null) {
                        mqIndex = Math.max(0, strZug.findOffset(sts) + 1);
                    } else {
                        String message = MessageFormat
                                .format("MQ {0} kann keinem Strassenteilsegment zugeordnet werden", mq);
                        applikation.getBetriebsmeldungssender().warnung(mq.getPid(), message);
                        LOGGER.warning(() -> message);
                    }
                    strZug.add(mqIndex, erzeugeIndikator(strZug, mq));
                }

                // MQ hinzufügen
                // for (final MessQuerschnittAllgemein mq : letztesSegment
                // .getMessquerschnitte()) {
                // final float aktuellerOffset = 0;
                // final float mqOffset = mq.getStrassenSegmentOffset();
                // boolean added = false;
                //
                // StrassenTeilSegment letztesSts = null;
                // for (final StrassenTeilSegment teilSegment : letztesSegment
                // .getStrassenTeilSegmente()) {
                // letztesSts = teilSegment;
                // if (mqOffset <= aktuellerOffset) {
                // final int tsIndex = Math.max(0, strZug
                // .indexOf(teilSegment));
                // strZug.add(tsIndex, erzeugeIndikator(strZug, mq));
                // added = true;
                // break;
                // }
                // }
                // if (!added) {
                // if (letztesSts != null) {
                // strZug.add(strZug.indexOf(letztesSts),
                // erzeugeIndikator(strZug, mq));
                // } else {
                // strZug.add(strZug.indexOf(letztesSegment),
                // erzeugeIndikator(strZug, mq));
                // }
                // }
                // }
                next = erzeugeIndikator(strZug, ((SobAeusseresStrassenSegment) next).getIndikator().getNachKnoten());
            } else if (next instanceof SobStrassenKnoten) {
                final SobStrassenKnoten knoten = (SobStrassenKnoten) next;
                next = erzeugeIndikator(strZug,
                        ((SobStrassenKnoten) next).getIndikator().getInnereVerbindungDanach(letztesSegment));
                knoten.setIss(next);
            } else if (next instanceof SobInneresStrassenSegment) {
                final int idx = strZug.indexOf(next) + 1;
                for (final StrassenTeilSegment teilSegment : ((SobInneresStrassenSegment) next).getIndikator()
                        .getStrassenTeilSegmente()) {
                    strZug.add(idx, erzeugeIndikator(strZug, teilSegment));
                }
                // if (StauverlaufsAnalyse.getOptionen().benutzeIssMq()) {

                for (final MessQuerschnittAllgemein mq : ((SobInneresStrassenSegment) next).getIndikator()
                        .getMessquerschnitte()) {
                    final StrassenTeilSegment sts = mq.getStrassenTeilSegment();
                    int mqIndex = strZug.indexOf(next) + 1;
                    if (sts != null) {
                        mqIndex = Math.max(0, strZug.findOffset(sts) + 1);
                    } else {
                        String message = MessageFormat
                                .format("MQ {0} kann keinem Strassenteilsegment zugeordnet werden", mq);
                        applikation.getBetriebsmeldungssender().warnung(mq.getPid(), message);
                        LOGGER.warning(() -> message);
                    }
                    strZug.add(mqIndex, erzeugeIndikator(strZug, mq));
                }
                // }
                AeusseresStrassenSegment nachSegment = ((SobInneresStrassenSegment) next).getIndikator()
                        .getNachSegment();
                if (!assListe.contains(nachSegment)) {
                    nachSegment = null;
                }
                next = erzeugeIndikator(strZug, nachSegment);
            }
        }
    }
}
