/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.bestimmungstau.parameter;

import java.util.Collection;
import java.util.TreeSet;
import java.util.WeakHashMap;
import java.util.concurrent.TimeUnit;

import de.bsvrz.dav.daf.main.ClientReceiverInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.Data.ReferenceArray;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.ReceiveOptions;
import de.bsvrz.dav.daf.main.ReceiverRole;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.VerkehrModellNetz;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Implementierung der Verwaltung von Parameterdaten für die
 * Stauobjektbestimmung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SobParameterObserver implements ISobParameterObserver, ClientReceiverInterface {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * Trennungszeichen für Attributname und Wert bei der Ausgabe der Parameter.
	 */
	private static final String DUMP_LABEL_TRENNER = ": ";

	/** der aktuelle Datensatz mit den Parameterdaten. */
	private Data parameterSet;

	/** die Attributgruppe mit den Parameterdaten. */
	private final AttributeGroup parameterAtg;

	/** die Liste der angemeldeten Listener. */
	private final WeakHashMap<ISobParameterListener, ISobParameterListener> listenerList = new WeakHashMap<>();

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt eine Instanz eines Parameterobservers für das
	 * angegebene VerkehsmodellNetz unter verwendung der übergebenen
	 * Datenverteilerverbindung.
	 *
	 * @param applikation
	 *            die Datenverteilerverbindung
	 * @param netz
	 *            das Objekt mit dem das Netz beschrieben wird
	 */
	public SobParameterObserver(final StauverlaufsAnalyse applikation, final VerkehrModellNetz netz) {

		final DataModel model = applikation.getDav().getDataModel();

		parameterAtg = model.getAttributeGroup(SobParameterDefs.ATG_NAME_STAUBESTIMMUNG_MODELL);
		final Aspect parameterAsp = model.getAspect("asp.parameterSoll");
		applikation.getDav().subscribeReceiver(this, netz.getSystemObject(), new DataDescription(parameterAtg, parameterAsp),
				ReceiveOptions.normal(), ReceiverRole.receiver());
	}

	@Override
	public final void addParameterListener(final ISobParameterListener listener) {
		synchronized (listenerList) {
			listenerList.put(listener, listener);
		}
	}

	@Override
	public final float getAnfangsLaenge() {
		float result = SobParameterDefs.SOB_DEFAULT_ANFANGSLAENGE;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_ANFANGSLAENGE).floatValue();
		}
		return result;
	}

	@Override
	public final float getMaxSiAbstand() {
		float result = SobParameterDefs.SOB_DEFAULT_MAX_SI_ABSTAND;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_MAX_SI_ABSTAND).floatValue();
		}
		return result;
	}

	@Override
	public final float getMinSiGuete() {
		float result = SobParameterDefs.SOB_DEFAULT_MIN_SI_GUETE;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_MIN_SI_GUETE).floatValue();
		}
		return result;
	}

	@Override
	public final String getParameterString() {
		final StringBuffer result = new StringBuffer();

		result.append(SobParameterDefs.ATT_ANFANGSLAENGE);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getAnfangsLaenge());
		result.append(SobParameterDefs.ATT_MAX_SI_ABSTAND);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getMaxSiAbstand());
		result.append(SobParameterDefs.ATT_MIN_SI_GUETE);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getMinSiGuete());
		result.append(SobParameterDefs.ATT_PROGNOSE_LAENGE_A1);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getPrognoseA1());
		result.append(SobParameterDefs.ATT_PROGNOSE_LAENGE_A2);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getPrognoseA2());
		result.append(SobParameterDefs.ATT_PROGNOSE_LAENGE_A3);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getPrognoseA3());
		result.append(SobParameterDefs.ATT_STROMAB_VERLAENGERUNG);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getStromabVerlaengerung());
		result.append(SobParameterDefs.ATT_TEILUNGSLAENGE);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getTeilungLaenge());
		result.append(SobParameterDefs.ATT_TEILUNGSVERZOEGERUNG);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getTeilungsVerzoegerung());
		result.append(SobParameterDefs.ATT_SI_VERFAHREN);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getVerfahren());
		result.append(SobParameterDefs.ATT_ZYKLUS_ZEIT);
		result.append(SobParameterObserver.DUMP_LABEL_TRENNER);
		result.append(getZyklusZeit());

		return result.toString();
	}

	@Override
	public final float getPrognoseA1() {
		float result = SobParameterDefs.SOB_DEFAULT_PROGNOSE_A1;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_PROGNOSE_LAENGE_A1).floatValue();
		}
		return result;
	}

	@Override
	public final float getPrognoseA2() {
		float result = SobParameterDefs.SOB_DEFAULT_PROGNOSE_A2;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_PROGNOSE_LAENGE_A2).floatValue();
		}
		return result;
	}

	@Override
	public final float getPrognoseA3() {
		float result = SobParameterDefs.SOB_DEFAULT_PROGNOSE_A3;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_PROGNOSE_LAENGE_A3).floatValue();
		}
		return result;
	}

	@Override
	public final float getStromabVerlaengerung() {
		float result = SobParameterDefs.SOB_DEFAULT_STROMAB_VERLAENGERUNG;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_STROMAB_VERLAENGERUNG).floatValue();
		}
		return result;
	}

	@Override
	public final float getTeilungLaenge() {
		float result = SobParameterDefs.SOB_DEFAULT_TEILUNGS_LAENGE;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_TEILUNGSLAENGE).floatValue();
		}
		return result;
	}

	@Override
	public final long getTeilungsVerzoegerung() {
		long result = SobParameterDefs.SOB_DEFAULT_TEILUNGS_VERZOEGERUNG;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_TEILUNGSVERZOEGERUNG).longValue();
		}
		return TimeUnit.SECONDS.toMillis(result);
	}

	@Override
	public final Collection<Aspect> getVerfahren() {
		final Collection<Aspect> result = new TreeSet<>();
		if (parameterSet != null) {
			final ReferenceArray array = parameterSet.getReferenceArray(SobParameterDefs.ATT_SI_VERFAHREN);
			for (int i = 0; i < array.getLength(); i++) {
				final SystemObject object = array.getSystemObject(i);
				if (object instanceof Aspect) {
					result.add((Aspect) object);
				}
			}
		}
		return result;
	}

	@Override
	public final long getZyklusZeit() {
		long result = SobParameterDefs.SOB_DEFAULT_ZYKLUS_IN_MINUTEN;
		if (parameterSet != null) {
			result = parameterSet.getScaledValue(SobParameterDefs.ATT_ZYKLUS_ZEIT).longValue();
		}
		return result;
	}

	@Override
	public final void removeParameterListener(final ISobParameterListener listener) {
		synchronized (listenerList) {
			listenerList.remove(listener);
		}
	}

	@Override
	public void update(final ResultData[] results) {
		boolean trigger = false;
		for (final ResultData result : results) {
			if (result.getDataDescription().getAttributeGroup().equals(parameterAtg)) {
				if (result.hasData()) {
					parameterSet = result.getData();
					SobParameterObserver.LOGGER.finer("SobParameter empfangen: " + parameterSet);
					trigger = true;
				} else {
					String message = "Keine Daten verfügbar";
					if (result.isNoSourceAvailable()) {
						message += ": Keine Quelle";
					} else if (result.isNoDataAvailable()) {
						message += ": Keine Daten";
					} else if (result.isNoRightsAvailable()) {
						message += ": Keine Rechte";
					}
					SobParameterObserver.LOGGER.info(message);
				}
			}
		}

		if (trigger) {
			synchronized (listenerList) {
				for (final ISobParameterListener listener : listenerList.values()) {
					listener.changed(this);
				}
			}
		}
	}
}
