/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.progstau;

import java.text.DateFormat;
import java.text.MessageFormat;
import java.time.Duration;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.TimeoutException;

import de.bsvrz.iav.gllib.gllib.domain.Intervall;
import de.bsvrz.iav.gllib.gllib.domain.MessQuerschnitt;
import de.bsvrz.iav.gllib.gllib.domain.Messwerte;
import de.bsvrz.iav.gllib.gllib.domain.MqGanglinie;
import de.bsvrz.iav.gllib.gllib.domain.PrognoseIntervall;
import de.bsvrz.iav.gllib.gllib.domain.comm.GlProgAntwort;
import de.bsvrz.iav.gllib.gllib.domain.comm.PrognoseAnfrage;
import de.bsvrz.ibv.anastau.anastau.StauverlaufsAnalyse;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobMessQuerschnitt;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStoerfallSituation;
import de.bsvrz.ibv.anastau.bestimmungstau.indikator.SobStrassenKnoten;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.IStau;
import de.bsvrz.ibv.anastau.bestimmungstau.stau.Stau;
import de.bsvrz.ibv.anastau.migration.modell.verkehr.objekte.MessQuerschnittAllgemein;
import de.bsvrz.ibv.anastau.migration.util.UndefiniertException;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zur Ermittlung der Zuflussmenge zu einem Stau.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SvpZuflussMenge implements ISvpZuflussMenge {

    /**
     * Logger für Debugausgaben.
     */
    private static final Debug LOGGER = Debug.getLogger();

    /**
     * die Ganglinie, die für die Ermittlung der Zuflussmenge genutzt werden soll.
     */
    private MqGanglinie gangLinie;

    /**
     * der Messquerschnitt, der die Daten für die Ermittlung der Zuflussmenge
     * liefern soll.
     */
    private MessQuerschnittAllgemein mq;

    /**
     * die Messquerschnitte auf Zufahrten vor dem Stau, die Daten für die Ermittlung
     * der Zuflussmenge liefern sollen.
     */
    private final Map<MessQuerschnittAllgemein, MqGanglinie> zufahrtenDavor = new HashMap<>();

    /**
     * die Messquerschnitte auf Abfahrten vor dem Stau, die Daten für die Ermittlung
     * der Zuflussmenge liefern sollen.
     */
    private final Map<MessQuerschnittAllgemein, MqGanglinie> abfahrtenDavor = new HashMap<>();

    /**
     * die Messquerschnitte auf Zufahrten im Stau, die Daten für die Ermittlung der
     * Zuflussmenge liefern sollen.
     */
    private final Map<MessQuerschnittAllgemein, MqGanglinie> zufahrtenImStau = new HashMap<>();

    /**
     * die aktuellen QB-Werte für Messquerschnitte auf Zu- und Abfahrten im Stau.
     */
    private final Map<MessQuerschnittAllgemein, Number> qbWerteMq = new HashMap<>();

    /**
     * die Messquerschnitte auf Abfahrten im Stau, die Daten für die Ermittlung der
     * Zuflussmenge liefern sollen.
     */
    private final Map<MessQuerschnittAllgemein, MqGanglinie> abfahrtenImStau = new HashMap<>();

    /**
     * der Stau für den der Zufluss ermittelt werden soll.
     */
    private final IStau stau;

    /** das Intervall für das die Zuflussmengen berechnet werden sollen. */
    private final Intervall intervall;

    /**
     * das Intervall das für die lineare Dämpfung der Ganglinien für Ab- und
     * Zufahrten im Stau verwendet wird. Zuflussdaten benötigt werden.
     */
    private final long daempfungsZeit;

    private StauverlaufsAnalyse applikation;

	private long schrittlaenge;

    /**
     * Konstruktor.<br>
     * Die Funktion erzeugt eine Instanz der Klasse zur Zuflussermittlung.
     *
     * @param applikation    die Applikation, die die Stauverlaufsanalyse durchführt
     * @param stau           der Stau, für den dr Zufluss ermittelt werden soll
     * @param intervall      das Intervall für das eine Prognose erfolgt und für das
     *                       Zuflussdaten benötigt werden.
     * @param daempfungsZeit das Intervall das für die lineare Dämpfung der
     *                       Ganglinien für Ab- und Zufahrten im Stau verwendet
     *                       wird. Zuflussdaten benötigt werden.
     * @param schrittlaenge  die Schrittlänge der Prognose, wird benötigt um potentielle Lücken auffüllen
     *                       zu können
     */
    public SvpZuflussMenge(StauverlaufsAnalyse applikation, final IStau stau, final Intervall intervall,
            final long daempfungsZeit, long schrittlaenge) {
        this.applikation = applikation;
        this.stau = stau;
		this.schrittlaenge = schrittlaenge;
        this.intervall = intervall;
        this.daempfungsZeit = daempfungsZeit;
        final SobMessQuerschnitt sobMq = ermittleMessquerschnitt();

        final List<PrognoseAnfrage> anfragen = new ArrayList<>();
        if (sobMq != null) {
            mq = sobMq.getIndikator();
            PrognoseIntervall prognoseIntervall = PrognoseIntervall.of(intervall.getStart(), intervall.getEnd());
            anfragen.add(new PrognoseAnfrage(MessQuerschnitt.of(mq.getPid()), prognoseIntervall, isMqImStau(sobMq)));

            if (StauverlaufsAnalyse.getOptionen().benutzeAnschluesseVorDemStau()) {
                for (final SobStrassenKnoten knoten : stau.ermittleAsAmEnde(mq)) {
                    for (final MessQuerschnittAllgemein mqObj : knoten.getAbfahrtMq()) {
                        abfahrtenDavor.put(mqObj, null);
                        anfragen.add(new PrognoseAnfrage(MessQuerschnitt.of(mqObj.getPid()), prognoseIntervall, true));
                    }
                    for (final MessQuerschnittAllgemein mqObj : knoten.getZufahrtMq()) {
                        zufahrtenDavor.put(mqObj, null);
                        anfragen.add(new PrognoseAnfrage(MessQuerschnitt.of(mqObj.getPid()), prognoseIntervall, true));
                    }
                }
            }

            if (StauverlaufsAnalyse.getOptionen().benutzeAnschluesseImStau()) {
                for (final SobStrassenKnoten knoten : stau.getKnotenImStau()) {
                    for (final MessQuerschnittAllgemein mqObj : knoten.getAbfahrtMq()) {
                        abfahrtenImStau.put(mqObj, null);
                        anfragen.add(new PrognoseAnfrage(MessQuerschnitt.of(mqObj.getPid()), prognoseIntervall, true));
                        qbWerteMq.put(mqObj, StauverlaufsPrognose.getMqQbWert(mqObj));
                    }
                    for (final MessQuerschnittAllgemein mqObj : knoten.getZufahrtMq()) {
                        zufahrtenImStau.put(mqObj, null);
                        anfragen.add(new PrognoseAnfrage(MessQuerschnitt.of(mqObj.getPid()), prognoseIntervall, true));
                        qbWerteMq.put(mqObj, StauverlaufsPrognose.getMqQbWert(mqObj));
                    }
                }
            }

            if (anfragen.size() > 0) {
                try {
                    GlProgAntwort antwort = applikation.getGanglinienprognose().sendeAnfrage(anfragen,
                            Duration.ofMinutes(2));
                    verarbeitePrognoseAntwort(antwort.getGanglinien());
                } catch (TimeoutException e) {
                    LOGGER.warning(() -> e.getLocalizedMessage());
                }
            }
        }
    }

    private final void verarbeitePrognoseAntwort(List<MqGanglinie> ganglinien) {

        SvpZuflussMenge.LOGGER.fine("Antwort von Ganglinienprognose eingetroffen");

        for (final MqGanglinie gl : ganglinien) {
            final MessQuerschnitt glMq = gl.getMessQuerschnitt();
            if (glMq == null) {
                continue;
            }

            if (glMq.getPid().equals(mq.getPid())) {
                gangLinie = gl;
            }

            for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(zufahrtenDavor.keySet())) {
                if (glMq.getPid().equals(mqObj.getPid())) {
                    zufahrtenDavor.put(mqObj, gl);
                    break;
                }
            }
            for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(abfahrtenDavor.keySet())) {
                if (glMq.getPid().equals(mqObj.getPid())) {
                    abfahrtenDavor.put(mqObj, gl);
                    break;
                }
            }
            for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(zufahrtenImStau.keySet())) {
                if (glMq.getPid().equals(mqObj.getPid())) {
                    zufahrtenImStau.put(mqObj, gl);
                    break;
                }
            }
            for (final MessQuerschnittAllgemein mqObj : new ArrayList<>(abfahrtenImStau.keySet())) {
                if (glMq.getPid().equals(mqObj.getPid())) {
                    abfahrtenImStau.put(mqObj, gl);
                    break;
                }
            }
        }
    }

    /**
     * ermittelt den Messquerschnitt, der für die Ermittlung des zufluss verwendet
     * wird.<br>
     * Es wird auf den ersten Messquerschnitt, der entgegengesetzt zur Fahrtrichtung
     * zwischen dem Stauende und dem nächsten Autobahnkreuz liegt zurückgegriffen.
     * Wird auf diese Art keiner gefunden, wird der letzte im Stau befindliche
     * Messquerschnitt eingesetzt.
     *
     * @return den zu verwendenden Messquerschnitt
     */
    private SobMessQuerschnitt ermittleMessquerschnitt() {
        SobMessQuerschnitt result = null;
        result = stau.getMqDavor();
        if (result == null) {
            result = stau.getMqAmEnde();
        }
        return result;
    }

    @Override
    public final double getZuflussMenge(final long zeitPunkt) throws UndefiniertException {
        if (gangLinie == null) {
            throw new UndefiniertException("Für MQ: \"" + mq + "\" konnte keine Ganglinie ermittelt werden");
        }

        Double wert = getZuflussAusGanglinie(gangLinie, zeitPunkt);
 
        // Zu- und Abfahrten vor dem Stau berücksichtigen

        for (final Entry<MessQuerschnittAllgemein, MqGanglinie> eintrag : zufahrtenDavor.entrySet()) {
            if (eintrag.getValue() != null) {
                Double zuflussWert = getZuflussAusGanglinie(eintrag.getValue(), zeitPunkt);
                
                if (zuflussWert == null) {
                    throw new UndefiniertException("Es existiert kein Wert für den Zeitpunkt");
                }
                wert = wert + zuflussWert;
            }
        }

        for (final Entry<MessQuerschnittAllgemein, MqGanglinie> eintrag : abfahrtenDavor.entrySet()) {
            if (eintrag.getValue() != null) {
                Double abflussWert = getZuflussAusGanglinie(eintrag.getValue(), zeitPunkt);
                wert = wert - abflussWert;
            }
        }

        // Zu- und Abfahrten im Stau berücksichtigen
        for (final Entry<MessQuerschnittAllgemein, MqGanglinie> eintrag : zufahrtenImStau.entrySet()) {
            final Number qbWert = qbWerteMq.get(eintrag.getKey());
            if (qbWert == null) {
                String message = MessageFormat.format(
                        "Für den im Stau liegenden {0} einer Zufahrt konnte kein aktueller QB-Wert ermittelt werden.",
                        eintrag.getKey());
                applikation.getBetriebsmeldungssender().warnung(eintrag.getKey().getSystemObject().getPid(), message);
                LOGGER.warning(() -> message);
            } else {
                if (eintrag.getValue() != null) {
                    Double anfangsWert = null;
					try {
						anfangsWert = getZuflussAusGanglinie(eintrag.getValue(), intervall.getStart());
					} catch (UndefiniertException e) {
						// hier ignoriert
					}
                    if (anfangsWert != null) {
                        Double zuflussWert = getZuflussAusGanglinie(eintrag.getValue(), zeitPunkt);
                        final double mqWert = zuflussWert.doubleValue()
                                + ((qbWert.doubleValue() - anfangsWert.doubleValue())
                                        * (1.0 - ((double) (zeitPunkt - intervall.getStart()) / daempfungsZeit)));
                        wert = wert + mqWert;
                    }
                }
            }
        }

        for (final Entry<MessQuerschnittAllgemein, MqGanglinie> eintrag : abfahrtenImStau.entrySet()) {
            final Number qbWert = qbWerteMq.get(eintrag.getKey());
            if (qbWert == null) {
                String message = MessageFormat.format(
                        "Für den im Stau liegenden {0} einer Abfahrt konnte kein aktueller QB-Wert ermittelt werden.",
                        eintrag.getKey());
                applikation.getBetriebsmeldungssender().warnung(eintrag.getKey().getSystemObject().getPid(), message);
                LOGGER.warning(() -> message);
            } else {
                if (eintrag.getValue() != null) {
                    Double anfangsWert = null;
					try {
						anfangsWert = getZuflussAusGanglinie(eintrag.getValue(), intervall.getStart());
					} catch (UndefiniertException e) {
						// hier ignorieren
					}
                    if (anfangsWert != null) {
                        Double zuflussWert = getZuflussAusGanglinie(eintrag.getValue(), zeitPunkt);
                        final double mqWert = zuflussWert.doubleValue()
                                + ((qbWert.doubleValue() - anfangsWert.doubleValue())
                                        * (1.0 - ((double) (zeitPunkt - intervall.getStart()) / daempfungsZeit)));
                        wert = wert - mqWert;
                    }
                }
            }
        }

        // Wert darf nicht kleiner als 0 sein.
        if (wert < 0) {
            wert = 0.0;
        }

        SvpZuflussMenge.LOGGER.finer("Zuflussmenge für MQ: " + mq + " um "
                + DateFormat.getDateTimeInstance().format(new Date(zeitPunkt)) + " = " + wert);
        return wert;
    }
    
	private Double getZuflussWertFor(MqGanglinie ganglinie, long zeitPunkt) {
		
		Messwerte messwerte = ganglinie.getStuetzstelle(zeitPunkt).getWert();
        if (messwerte == null ) {
        	return null;
        } 
        
        return messwerte.getQB();
	}

    private Double getZuflussAusGanglinie(MqGanglinie ganglinie, long zeitPunkt) throws UndefiniertException {
 
    	Double zufluss = getZuflussWertFor(ganglinie, zeitPunkt);
        if (zufluss == null) {
        	return interpoliereZufluss(ganglinie, zeitPunkt, applikation.getMaxGapSize());
        } 
        
        return zufluss;
 	}



	private Double interpoliereZufluss(MqGanglinie ganglinie, long zeitPunkt, int maxGapSize) throws UndefiniertException {
		if( maxGapSize <= 0 || schrittlaenge <= 0) {
			throw new UndefiniertException("Kein Messwert für den Zeitpunkt " + DateFormat.getDateTimeInstance().format(new Date(zeitPunkt)) + " verfügbar");
		}
		
		int gapsVorher = 1;
		Double wertVorher = null;
		long zs = zeitPunkt - schrittlaenge;
		while(zs >= intervall.getStart() && wertVorher == null) {
			wertVorher = getZuflussWertFor(ganglinie, zs);
			if(wertVorher == null) {
				gapsVorher++;
			}
			zs -= schrittlaenge;
		}
		
		Double wertDanach = null;
		int gapsDanach = 0;
		zs = zeitPunkt + schrittlaenge;
		while(zs <= intervall.getEnd() && wertDanach == null) {
			wertDanach = getZuflussWertFor(ganglinie, zs);
			if(wertDanach == null) {
				gapsDanach++;
			}
			zs += schrittlaenge;
		}
		
		int gaps = gapsVorher + gapsDanach;
		
		if (gaps > maxGapSize || (wertVorher == null && wertDanach == null)) {
			throw new UndefiniertException("Kein Messwert für den aktuellen Zeitpunkt verfügbar");
		}
		
		if( wertVorher == null) {
			return wertDanach;
		}

		if( wertDanach == null) {
			return wertVorher;
		}

		double interpolationsSchritt = (wertDanach - wertVorher) / (gaps + 1);
		return wertVorher + (gapsVorher * interpolationsSchritt);
	}

	/**
     * ermittelt, ob der übergebene Messquerschnitt innerhalb des Staus der
     * Zuflussmengenbestimmung liegt.
     *
     * @param mqObj das MQ-Objekt
     * @return true, wenn der MQ im Stau liegt
     */
    private boolean isMqImStau(final SobMessQuerschnitt mqObj) {
        boolean result = false;

        if (mqObj.getSituation() == SobStoerfallSituation.STAU) {
            result = true;
        } else {
            for (final Stau aktuellerStau : mqObj.getRichtung().getStaus().getStauMenge()) {

                final double stauBeginn = aktuellerStau.getStrassenStart();
                final double stauEnde = aktuellerStau.getStrassenEnde();
                final double position = mqObj.getPosition();

                if ((position > stauBeginn) && (position < stauEnde)) {
                    result = true;
                    break;
                }
            }
        }

        return result;
    }
}
